/*******************************************************************\

Module:

Author: Daniel Kroening, kroening@cs.cmu.edu

\*******************************************************************/

#ifndef CPROVER_C_TYPES_H
#define CPROVER_C_TYPES_H

#include <expr.h>

typet index_type();
typet enum_type();
typet int_type();
typet uint_type();
typet long_int_type();
typet long_long_int_type();
typet long_uint_type();
typet long_long_uint_type();
typet char_type();
typet float_type();
typet double_type();
typet long_double_type();

#endif
